==> klayout-0.19.3.tar.gz: klayout-0.19.3/src/layPlugin.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layPlugin.h"

#include "tlClassRegistry.h"
#include "tlException.h"
#include "tlAssert.h"
#include "tlXMLParser.h"

#include <iostream>
#include <fstream>

//  The instance of the PluginDeclaration class registrar
template <> tl::Registrar<lay::PluginDeclaration> *tl::Registrar<lay::PluginDeclaration>::instance = 0;

namespace lay
{

// ----------------------------------------------------------------
//  Plugin implementation

Plugin::Plugin (Plugin *parent)
  : mp_parent (parent), mp_plugin_declaration (0), dm_finalize_config (this, &lay::Plugin::config_end)
{
  if (! parent) {
    //  load the root with the default configuration
    for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
      std::vector< std::pair<std::string, std::string> > pairs;
      cls->get_options (pairs);
      m_repository.insert (pairs.begin (), pairs.end ());
    }
  } else {
    mp_parent->m_children.insert (this, false /*not owned*/);
  }
}

Plugin::~Plugin ()
{
  // .. nothing yet ..
}

void 
Plugin::config_setup ()
{
  do_config_setup (this);
  do_config_end ();
}
  
void 
Plugin::config_set (const std::string &name, const std::string &value)
{
  //  install the value in the repository
  m_repository.insert (std::make_pair (name, std::string ())).first->second = value;

  //  look for plugins that receive that configuration statically if the root is addressed
  if (! mp_parent) {
    for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
      if ((const_cast<lay::PluginDeclaration *> (&*cls))->configure (name, value)) {
        return;
      }
    }
  }

  do_config_set (name, value);

  //  schedule a configuration finalization call (once for all config_set calls)
  dm_finalize_config ();
}

void 
Plugin::config_end ()
{
  //  finish static conifgurations for plugins if the root is addressed
  if (! mp_parent) {
    for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
      (const_cast<lay::PluginDeclaration *> (&*cls))->config_finalize ();
    }
  }

  do_config_end ();
}

void
Plugin::clear_config ()
{
  m_repository.clear ();
  if (! mp_parent) {
    //  load the root with the default configuration
    for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
      std::vector< std::pair<std::string, std::string> > pairs;
      cls->get_options (pairs);
      m_repository.insert (pairs.begin (), pairs.end ());
    }
  }
  do_config_end ();
}

bool 
Plugin::config_get (const std::string &name, std::string &value) const
{
  std::map<std::string, std::string>::const_iterator p = m_repository.find (name);
  if (p != m_repository.end ()) {
    value = p->second;
    return true;
  } else {
    value = "";
    return false;
  }
}

void 
Plugin::get_config_names (std::vector<std::string> &names)
{
  //  retrieve the configuration names
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {

    std::vector< std::pair<std::string, std::string> > pairs;
    cls->get_options (pairs);
    for (std::vector< std::pair<std::string, std::string> >::const_iterator p = pairs.begin (); p != pairs.end (); ++p) {
      names.push_back (p->first);
    }

  }
}

void 
Plugin::do_config_setup (Plugin *target)
{
  if (mp_parent) {
    mp_parent->do_config_setup (target);
  } 
  //  local configurations override the parent's configuration, i.e. are applied after the parents settings
  for (std::map<std::string, std::string>::const_iterator p = m_repository.begin (); p != m_repository.end (); ++p) {
    target->do_config_set (p->first, p->second);
  }
}

void 
Plugin::do_config_end ()
{
  config_finalize ();
  for (tl::Collection<Plugin>::iterator c = m_children.begin (); c != m_children.end (); ++c) {
    c->do_config_end ();
  }
}

bool 
Plugin::do_config_set (const std::string &name, const std::string &value)
{
  if (configure (name, value)) {
    //  taken by us - don't propagate to the children
    return true;
  }

  //  propagate to all children (not only the first that takes it!)
  for (tl::Collection<Plugin>::iterator c = m_children.begin (); c != m_children.end (); ++c) {
    c->do_config_set (name, value);
  }

  return false;
}

// ----------------------------------------------------------------
//  PluginRoot implementation

//  Writing and Reading of configuration 

struct ConfigGetAdaptor
{
  ConfigGetAdaptor (const std::string &name)
    : mp_owner (0), m_name (name)
  {
    // .. nothing yet ..
  }
  
  std::string operator () () const
  {
    std::string s;
    mp_owner->config_get (m_name, s);
    return s;
  }

  bool at_end () const 
  {
    return m_done;
  }

  void start (const lay::PluginRoot &owner) 
  {
    mp_owner = &owner;
    m_done = false;
  }

  void next () 
  {
    m_done = true;
  }

private:
  const lay::PluginRoot *mp_owner;
  bool m_done;
  std::string m_name;
};
 
struct ConfigSetAdaptor
{
  ConfigSetAdaptor (const std::string &name)
    : m_name (name)
  {
    // .. nothing yet ..
  }
  
  void operator () (lay::PluginRoot &w, tl::XMLReaderState &reader) const
  {
    tl::XMLObjTag<std::string> tag;
    w.config_set (m_name, *reader.back (tag));
  }

private:
  std::string m_name;
};
 
//  the configuration file's XML structure is built dynamically
static tl::XMLStruct<lay::PluginRoot> 
config_structure ()
{
  tl::XMLElementList body;

  std::vector <std::string> names; 
  lay::Plugin::get_config_names (names);
  for (std::vector <std::string>::const_iterator n = names.begin (); n != names.end (); ++n) {
    body.append (tl::XMLMember<std::string, lay::PluginRoot, ConfigGetAdaptor, ConfigSetAdaptor, tl::XMLStdConverter <std::string> > ( 
                       ConfigGetAdaptor (*n), ConfigSetAdaptor (*n), *n));
  }

  //  add a wildcard member to ignore all others
  body.append (tl::make_member<lay::PluginRoot> ("*"));

  return tl::XMLStruct<lay::PluginRoot> ("config", body);
}


bool 
PluginRoot::write_config (const std::string &config_file)
{
  std::ofstream out (config_file.c_str ());
  if (! out.good ()) {
    return false;
  }
  config_structure ().write (out, *this); 
  return true;
}

bool
PluginRoot::read_config (const std::string &config_file)
{
  bool ret = false;

  try {
    tl::XMLFileSource file (config_file);
    try {
      //  clear the custom dither pattern
      config_structure ().parse (file, *this); 
      ret = true;
    } catch (tl::XMLException &ex) {
      tl::error << tl::translate ("Problem reading config file:") << ex.msg ();
    } 
  } catch (tl::Exception &) {
    // .. ignore errors ..
  }

  config_end ();
  return ret;
}

// ----------------------------------------------------------------
//  ColorConverter implementation

std::string 
ColorConverter::to_string (const QColor &c)
{
  if (! c.isValid ()) {
    return "auto";
  } else {
    return c.name ().ascii ();
  }
}

void 
ColorConverter::from_string (const std::string &s, QColor &c)
{
  std::string t (tl::trim (s));
  if (t == "auto") {
    c = QColor ();
  } else {
    c = QColor (t.c_str ());
  } 
}

}

==> klayout-0.21.3.tar.gz: klayout-0.21.3/src/layPlugin.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layPlugin.h"

#include "tlClassRegistry.h"
#include "tlException.h"
#include "tlAssert.h"
#include "tlXMLParser.h"

#include <iostream>
#include <fstream>
#include <memory>

//  The instance of the PluginDeclaration class registrar
template <> tl::Registrar<lay::PluginDeclaration> *tl::Registrar<lay::PluginDeclaration>::instance = 0;

namespace lay
{

// ----------------------------------------------------------------
//  Plugin implementation

Plugin::Plugin (Plugin *parent, bool standalone)
  : mp_parent (parent), mp_plugin_declaration (0), dm_finalize_config (this, &lay::Plugin::config_end), m_standalone (standalone)
{
  if (! parent) {
    if (! standalone) {
      //  load the root with the default configuration
      for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
        std::vector< std::pair<std::string, std::string> > pairs;
        cls->get_options (pairs);
        m_repository.insert (pairs.begin (), pairs.end ());
      }
    }
  } else {
    mp_parent->m_children.insert (this, false /*not owned*/);
  }
}

Plugin::~Plugin ()
{
  // .. nothing yet ..
}

void 
Plugin::config_setup ()
{
  do_config_setup (this);
  do_config_end ();
}
  
void 
Plugin::config_set (const std::string &name, const std::string &value)
{
  std::map <std::string, std::string>::iterator m = m_repository.find (name);
  if (m != m_repository.end ()) {
    //  if the value did not change, do nothing 
    if (m->second == value) {
      return;
    } 
    m->second = value;
  } else {
    //  install the value in the repository
    m_repository.insert (std::make_pair (name, value));
  } 

  //  look for plugins that receive that configuration statically if the root is addressed
  if (! mp_parent && ! m_standalone) {
    for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
      if ((const_cast<lay::PluginDeclaration *> (&*cls))->configure (name, value)) {
        return;
      }
    }
  }

  do_config_set (name, value);

  //  schedule a configuration finalization call (once for all config_set calls)
  dm_finalize_config ();
}

void 
Plugin::config_end ()
{
  //  finish static conifgurations for plugins if the root is addressed
  if (! mp_parent && ! m_standalone) {
    for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
      (const_cast<lay::PluginDeclaration *> (&*cls))->config_finalize ();
    }
  }

  do_config_end ();
}

void
Plugin::clear_config ()
{
  m_repository.clear ();
  if (! mp_parent && ! m_standalone) {
    //  load the root with the default configuration
    for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
      std::vector< std::pair<std::string, std::string> > pairs;
      cls->get_options (pairs);
      m_repository.insert (pairs.begin (), pairs.end ());
    }
  }
  do_config_end ();
}

bool 
Plugin::config_get (const std::string &name, std::string &value) const
{
  std::map<std::string, std::string>::const_iterator p = m_repository.find (name);
  if (p != m_repository.end ()) {
    value = p->second;
    return true;
  } else {
    value = "";
    return false;
  }
}

void 
Plugin::get_config_names (std::vector<std::string> &names)
{
  //  retrieve the configuration names
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {

    std::vector< std::pair<std::string, std::string> > pairs;
    cls->get_options (pairs);
    for (std::vector< std::pair<std::string, std::string> >::const_iterator p = pairs.begin (); p != pairs.end (); ++p) {
      names.push_back (p->first);
    }

  }
}

void 
Plugin::do_config_setup (Plugin *target)
{
  if (mp_parent) {
    mp_parent->do_config_setup (target);
  } 
  //  local configurations override the parent's configuration, i.e. are applied after the parents settings
  for (std::map<std::string, std::string>::const_iterator p = m_repository.begin (); p != m_repository.end (); ++p) {
    target->do_config_set (p->first, p->second);
  }
}

void 
Plugin::do_config_end ()
{
  config_finalize ();
  for (tl::Collection<Plugin>::iterator c = m_children.begin (); c != m_children.end (); ++c) {
    c->do_config_end ();
  }
}

bool 
Plugin::do_config_set (const std::string &name, const std::string &value)
{
  if (configure (name, value)) {
    //  taken by us - don't propagate to the children
    return true;
  }

  //  propagate to all children (not only the first that takes it!)
  for (tl::Collection<Plugin>::iterator c = m_children.begin (); c != m_children.end (); ++c) {
    c->do_config_set (name, value);
  }

  return false;
}

// ----------------------------------------------------------------
//  PluginRoot implementation

//  Writing and Reading of configuration 

struct ConfigGetAdaptor
{
  ConfigGetAdaptor (const std::string &name)
    : mp_owner (0), m_name (name)
  {
    // .. nothing yet ..
  }
  
  std::string operator () () const
  {
    std::string s;
    mp_owner->config_get (m_name, s);
    return s;
  }

  bool at_end () const 
  {
    return m_done;
  }

  void start (const lay::PluginRoot &owner) 
  {
    mp_owner = &owner;
    m_done = false;
  }

  void next () 
  {
    m_done = true;
  }

private:
  const lay::PluginRoot *mp_owner;
  bool m_done;
  std::string m_name;
};
 
struct ConfigGetNullAdaptor
{
  ConfigGetNullAdaptor ()
  {
    // .. nothing yet ..
  }
  
  std::string operator () () const
  {
    return std::string ();
  }

  bool at_end () const 
  {
    return true;
  }

  void start (const lay::PluginRoot &owner) { }
  void next () { }
};
 
struct ConfigSetAdaptor
{
  ConfigSetAdaptor (const std::string &name)
    : m_name (name)
  {
    // .. nothing yet ..
  }
  
  void operator () (lay::PluginRoot &w, tl::XMLReaderState &reader) const
  {
    tl::XMLObjTag<std::string> tag;
    w.config_set (m_name, *reader.back (tag));
  }

private:
  std::string m_name;
};
 
//  the configuration file's XML structure is built dynamically
static tl::XMLStruct<lay::PluginRoot> 
config_structure ()
{
  tl::XMLElementList body;
  std::string n_with_underscores;

  std::vector <std::string> names; 
  lay::Plugin::get_config_names (names);

  for (std::vector <std::string>::const_iterator n = names.begin (); n != names.end (); ++n) {

    body.append (tl::XMLMember<std::string, lay::PluginRoot, ConfigGetAdaptor, ConfigSetAdaptor, tl::XMLStdConverter <std::string> > ( 
                       ConfigGetAdaptor (*n), ConfigSetAdaptor (*n), *n));

    //  for compatibility, provide an alternative with underscores (i.e. 0.20->0.21 because of default_grids)
    n_with_underscores.clear ();
    for (const char *c = n->c_str (); *c; ++c) {
      n_with_underscores += (*c == '-' ? '_' : *c);
    }

    body.append (tl::XMLMember<std::string, lay::PluginRoot, ConfigGetNullAdaptor, ConfigSetAdaptor, tl::XMLStdConverter <std::string> > ( 
                       ConfigGetNullAdaptor (), ConfigSetAdaptor (*n), n_with_underscores));

  }

  //  add a wildcard member to ignore all others
  body.append (tl::make_member<lay::PluginRoot> ("*"));

  return tl::XMLStruct<lay::PluginRoot> ("config", body);
}


bool 
PluginRoot::write_config (const std::string &config_file)
{
  std::ofstream out (config_file.c_str ());
  if (! out.good ()) {
    return false;
  }
  config_structure ().write (out, *this); 
  return true;
}

bool
PluginRoot::read_config (const std::string &config_file)
{
  std::auto_ptr<tl::XMLFileSource> file;

  try {
    file.reset (new tl::XMLFileSource (config_file));
  } catch (...) {
    return false;
  }

  try {
    config_structure ().parse (*file, *this); 
  } catch (tl::Exception &ex) {
    std::string msg = tl::translate ("Problem reading config file ") + config_file + ": " + ex.msg ();
    throw tl::Exception (msg);
  } 

  config_end ();

  return true;
}

}
